from datetime import timedelta
from pytest import mark

from django.core import mail
from django.shortcuts import reverse
from django.utils import timezone

from magicauth.models import MagicToken
from tests import factories

pytestmark = mark.django_db


def test_getting_LoginView_while_authenticated_redirects_to_default(client):
    user = factories.UserFactory()
    client.force_login(user)
    url = reverse("magicauth-login")
    response = client.get(url)
    assert response.status_code == 302
    assert response.url == "/test_home/"


def test_getting_LoginView_while_authenticated_with_next_redirects_to_next(client):
    user = factories.UserFactory()
    client.force_login(user)
    url = reverse("magicauth-login") + "?next=/test_dashboard/"
    response = client.get(url)
    assert response.status_code == 302
    assert response.url == "/test_dashboard/"


def test_posting_email_for_valid_existing_user_redirects(client):
    user = factories.UserFactory()
    url = reverse("magicauth-login")
    data = {"email": user.email}
    response = client.post(url, data=data)
    assert response.status_code == 302
    assert len(mail.outbox) == 1


def test_loging_with_email_is_case_insensitive(client):
    user = factories.UserFactory()
    url = reverse("magicauth-login")
    data = {"email": user.email.upper()}
    response = client.post(url, data=data)
    assert response.status_code == 302
    assert len(mail.outbox) == 1


def test_posting_unknown_email_raise_error(client):
    url = reverse("magicauth-login")
    data = {"email": "unknown@email.com"}
    response = client.post(url, data=data)
    assert "invalid" in str(response.content)
    assert len(mail.outbox) == 0


def test_posting_email_for_valid_existing_user_sends_email(client):
    user = factories.UserFactory()
    url = reverse("magicauth-login")
    data = {"email": user.email}
    client.post(url, data=data)
    assert len(mail.outbox) == 1
    assert "?next=/test_home/" in mail.outbox[0].body


def test_posting_email_redirect_to_default_view(client):
    user = factories.UserFactory()
    url = reverse("magicauth-login")
    data = {"email": user.email}
    client.post(url, data=data)
    assert "?next=/test_home/" in mail.outbox[0].body


def test_posting_email_with_next_redirects_to_next(client):
    user = factories.UserFactory()
    url = reverse("magicauth-login") + "?next=/test_dashboard/"
    data = {"email": user.email}
    client.post(url, data=data)
    assert "?next=/test_dashboard/" in mail.outbox[0].body


def test_posting_unknown_email_does_not_send_email(client):
    url = reverse("magicauth-login")
    data = {"email": "unknown@email.com"}
    client.post(url, data=data)
    assert len(mail.outbox) == 0


def test_posting_email_for_valid_existing_user_created_token(client):
    user = factories.UserFactory()
    url = reverse("magicauth-login")
    data = {"email": user.email}
    count_before = MagicToken.objects.count()
    client.post(url, data=data)
    count_after = MagicToken.objects.count()
    assert count_after == count_before + 1


def test_opening_magic_link_with_valid_token_redirects(client):
    token = factories.MagicTokenFactory()
    url = reverse("magicauth-validate-token", args=[token.key])
    response = client.get(url)
    assert response.status_code == 302
    assert response.url == "/test_home/"


def test_opening_magic_link_with_a_next_sets_a_new_url(client):
    token = factories.MagicTokenFactory()
    url = (
        reverse("magicauth-validate-token", kwargs={"key": token.key})
        + "?next=/test_dashboard/?a=test&b=test"
    )
    response = client.get(url)
    assert response.status_code == 302
    assert response.url == "/test_dashboard/?a=test&b=test"


def test_opening_magic_link_with_a_unsafe_next_sets_triggers_404(client):
    token = factories.MagicTokenFactory()
    url = (
        reverse("magicauth-validate-token", kwargs={"key": token.key})
        + "?next=http://www.myfishingsite.com/?a=test&b=test"
    )
    response = client.get(url)
    assert response.status_code == 404


def test_opening_magic_link_with_a_unsafe_next_while_loggedin_sets_triggers_404(client):
    token = factories.MagicTokenFactory()
    user = factories.UserFactory()
    client.force_login(user)
    url = (
        reverse("magicauth-validate-token", kwargs={"key": token.key})
        + "?next=http://www.myfishingsite.com/?a=test&b=test"
    )
    response = client.get(url)
    assert response.status_code == 404
    assert user.is_authenticated == True


def test_token_is_removed_after_visiting_magic_link(client):
    token = factories.MagicTokenFactory()
    url = reverse("magicauth-validate-token", args=[token.key])
    count_before = MagicToken.objects.count()
    client.get(url)
    count_after = MagicToken.objects.count()
    assert count_after == count_before - 1


def test_duplicate_token_for_same_user_is_removed_after_visiting_magic_link(client):
    token = factories.MagicTokenFactory()
    duplicate = factories.MagicTokenFactory(user=token.user)
    url = reverse("magicauth-validate-token", args=[token.key])
    client.get(url)
    assert duplicate not in MagicToken.objects.all()


def test_visiting_magic_link_triggers_login(client):
    token = factories.MagicTokenFactory()
    url = reverse("magicauth-validate-token", args=[token.key])
    client.get(url)
    assert "_auth_user_id" in client.session


def test_unknown_token_redirects(client):
    url = reverse("magicauth-validate-token", args=["unknown-token"])
    response = client.get(url)
    assert response.status_code == 302


def test_expired_token_redirects(client):
    token = factories.MagicTokenFactory()
    token.created = timezone.now() - timedelta(days=1)
    token.save()
    url = reverse("magicauth-validate-token", args=[token.key])
    response = client.get(url)
    assert response.status_code == 302


def test_expired_token_is_deleted(client):
    token = factories.MagicTokenFactory()
    token.created = timezone.now() - timedelta(days=1)
    token.save()
    url = reverse("magicauth-validate-token", args=[token.key])
    client.get(url)
    assert token not in MagicToken.objects.all()



from django.conf import settings as django_settings

# To add magicauth to your site, you need to add these values to your site settings
# (the rest have defaults, you are free to change them if you want):
# MAGICAUTH_FROM_EMAIL : e.g. 'contact@mysite.com'
# MAGICAUTH_LOGGED_IN_REDIRECT_URL_NAME : e.g. 'home'
#
# Note : for the default templates to be picked up, you need to enable
# the app_directories template Loader, by setting TEMPLATES app_dirs to True.
# https://docs.djangoproject.com/en/2.2/ref/templates/api/#django.template.loaders.app_directories.Loader

# Note 2 : magicauth assumes your logout url name is 'logout'.
# If it's not, change MAGICAUTH_LOGOUT_URL_NAME in your settings.

# Email settings
EMAIL_SUBJECT = getattr(django_settings, "MAGICAUTH_EMAIL_SUBJECT", "Lien de connexion")
EMAIL_HTML_TEMPLATE = getattr(
    django_settings, "MAGICAUTH_EMAIL_HTML_TEMPLATE", "magicauth/email.html"
)
EMAIL_TEXT_TEMPLATE = getattr(
    django_settings, "MAGICAUTH_EMAIL_TEXT_TEMPLATE", "magicauth/email.txt"
)
FROM_EMAIL = getattr(django_settings, "MAGICAUTH_FROM_EMAIL")

# View templates
LOGIN_VIEW_TEMPLATE = getattr(
    django_settings, "MAGICAUTH_LOGIN_VIEW_TEMPLATE", "magicauth/login.html"
)
EMAIL_SENT_VIEW_TEMPLATE = getattr(
    django_settings, "MAGICAUTH_EMAIL_SENT_VIEW_TEMPLATE", "magicauth/email_sent.html"
)

# URLs for magicauth views
# Once user has entered email successfully and email has been sent, show this page.
EMAIL_SENT_URL = getattr(django_settings, "MAGICAUTH_EMAIL_SENT_URL", "email-envoyé/")
LOGIN_URL = getattr(django_settings, "MAGICAUTH_LOGIN_URL", "login/")
# The emailed links point to this url.
# The view will look for the token in the "key" variable.
VALIDATE_TOKEN_URL = getattr(
    django_settings, "MAGICAUTH_VALIDATE_TOKEN_URL", "code/<str:key>/"
)

# URL names for hooking up magicauth to your site
# Once user is logged in, redirect to this url (probably your landing page).
LOGGED_IN_REDIRECT_URL_NAME = getattr(
    django_settings, "MAGICAUTH_LOGGED_IN_REDIRECT_URL_NAME"
)
LOGOUT_URL_NAME = getattr(django_settings, "MAGICAUTH_LOGOUT_URL_NAME", "logout")

# Name of the field in your User model that contains the email
EMAIL_FIELD = getattr(django_settings, "MAGICAUTH_EMAIL_FIELD", "username")

# Other
# How long a token stays valid.
# When using an expired token, user will be prompted to get a new one.
TOKEN_DURATION_SECONDS = getattr(
    django_settings, "MAGICAUTH_TOKEN_DURATION_SECONDS", 5 * 60
)
# Function to call when the email entered in the form is not found in the database.
# The default just raises an error whose message gets displayed on the login page.
EMAIL_UNKNOWN_CALLBACK = getattr(
    django_settings, "MAGICAUTH_EMAIL_UNKNOWN_CALLBACK", "magicauth.utils.raise_error"
)
# If using the default EMAIL_UNKNOWN_CALLBACK,
# this message will be displayed when an unknown email is entered.
EMAIL_UNKNOWN_MESSAGE = getattr(
    django_settings, "MAGICAUTH_EMAIL_UNKNOWN_MESSAGE", "Aucun utilisateur trouvé."
)



import re
import logging
from datetime import timedelta

from django.contrib import messages
from django.contrib.auth import login
from django.http import HttpResponseNotFound
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils import timezone
from django.utils.http import is_safe_url
from django.views.generic import View, FormView, TemplateView

from magicauth.forms import EmailForm
from magicauth.models import MagicToken
from magicauth import settings as magicauth_settings

logger = logging.getLogger()

class LoginView(FormView):
    """
    The login page. The user enters their email in the form to get a link by email.
    """

    form_class = EmailForm
    success_url = reverse_lazy("magicauth-email-sent")
    template_name = magicauth_settings.LOGIN_VIEW_TEMPLATE

    def get(self, request, *args, **kwargs):
        next_view = self.request.GET.get(
            "next", f"/{magicauth_settings.LOGGED_IN_REDIRECT_URL_NAME}/"
        )
        if request.user.is_authenticated:
            return redirect(next_view)

        return super(LoginView, self).get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super(LoginView, self).get_context_data(**kwargs)
        context[
            "LOGGED_IN_REDIRECT_URL_NAME"
        ] = magicauth_settings.LOGGED_IN_REDIRECT_URL_NAME
        context["LOGOUT_URL_NAME"] = magicauth_settings.LOGOUT_URL_NAME
        return context

    def form_valid(self, form, *args, **kwargs):
        next_view = self.request.GET.get(
            "next",
            f"/{magicauth_settings.LOGGED_IN_REDIRECT_URL_NAME}/"
        )
        current_site = self.request.site
        form.send_email(current_site, next_view)
        return super().form_valid(form)


class EmailSentView(TemplateView):
    """
    View shown to confirm the email has been sent.
    """

    template_name = magicauth_settings.EMAIL_SENT_VIEW_TEMPLATE


class ValidateTokenView(View):
    """
    The link sent by email goes to this view.
    It validates the token passed in querystring,
    and either logs in or shows a form to make a new token.
    """

    @staticmethod
    def get_valid_token(key):
        duration = magicauth_settings.TOKEN_DURATION_SECONDS
        try:
            token = MagicToken.objects.get(key=key)
        except MagicToken.DoesNotExist:
            return None
        except MagicToken.MultipleObjectsReturned:
            return None

        if token.created < timezone.now() - timedelta(seconds=duration):
            token.delete()
            return None
        return token

    def get(self, request, *args, **kwargs):
        full_path = request.get_full_path()

        rule_for_redirect = re.compile("(.*next=)(.*)")
        next_view = rule_for_redirect.match(full_path)
        redirect_default = reverse_lazy(magicauth_settings.LOGGED_IN_REDIRECT_URL_NAME)
        url = next_view.group(2) if next_view else redirect_default

        # the following `is_safe_url` will be deprecated in django 4 and replaced by url_has_allowed_host_and_scheme
        if not is_safe_url(url, allowed_hosts={request.get_host()}, require_https=True):
            # We are not logging the unsafe URL to prevent code injections in logs
            logger.warning("[MagicAuth] an unsafe URL was used through a login link")
            return HttpResponseNotFound()

        if request.user.is_authenticated:
            return redirect(url)
        token_key = kwargs.get("key")
        token = self.get_valid_token(token_key)
        if not token:
            messages.warning(
                self.request,
                "Ce lien de connexion ne fonctionne plus. "
                "Pour en recevoir un nouveau, nous vous invitons à renseigner "
                "votre email ci-dessous puis à cliquer sur valider.",
            )
            return redirect("magicauth-login")
        login(self.request, token.user)
        MagicToken.objects.filter(
            user=token.user
        ).delete()  # Remove them all for this user
        return redirect(url)
